#!/usr/bin/env python
#
# Error types for pymwalib
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#

class PymwalibError(Exception):
    """Base class for other exceptions"""
    pass

class PymwalibMwalibVersionNotCompatibleError(PymwalibError):
    """Raised when mwalib is not a compatible version"""
    pass

class PymwalibMetafitsContextNewError(PymwalibError):
    """Raised when call to C mwalib_metafits_context_new fails"""
    pass

class PymwalibCorrelatorContextNewError(PymwalibError):
    """Raised when call to C mwalib_correaltor_context_new fails"""
    pass

class PymwalibVoltageContextNewError(PymwalibError):
    """Raised when call to C mwalib_voltage_context_new fails"""
    pass

class PymwalibAntennasGetError(PymwalibError):
    """Raised when call to C mwalib_antennas_get fails"""
    pass

class PymwalibBaselinesGetError(PymwalibError):
    """Raised when call to C mwalib_baselines_get fails"""
    pass

class PymwalibCoarseChannelsGetError(PymwalibError):
    """Raised when call to C mwalib_coarse_channels_get fails"""
    pass

class PymwalibMetafitsMetadataGetError(PymwalibError):
    """Raised when call to C mwalib_metafits_metadata_get fails"""
    pass

class PymwalibCorrelatorMetadataGetError(PymwalibError):
    """Raised when call to C mwalib_correlator_metadata_get fails"""
    pass

class PymwalibVoltageMetadataGetError(PymwalibError):
    """Raised when call to C mwalib_voltage_metadata_get fails"""
    pass

class PymwalibRFInputsGetError(PymwalibError):
    """Raised when call to C mwalib_rfinputs_get fails"""
    pass

class PymwalibTimeStepsGetError(PymwalibError):
    """Raised when call to C mwalib_timesteps_get fails"""
    pass

class PymwalibMetafitsContextDisplayError(PymwalibError):
    """Raised when call to C mwalib_metafits_context_display fails"""
    pass

class PymwalibCorrelatorContextDisplayError(PymwalibError):
    """Raised when call to C mwalib_correlator_context_display fails"""
    pass

class PymwalibVoltageContextDisplayError(PymwalibError):
    """Raised when call to C mwalib_voltage_context_display fails"""
    pass

class PymwalibVoltageContextReadFileError(PymwalibError):
    """Raised when call to C mwalib_voltage_context_read_file fails"""
    pass

class PymwalibVoltageContextReadSecondError(PymwalibError):
    """Raised when call to C mwalib_voltage_context_read_second fails"""
    pass

class PymwalibCorrelatorContextReadByBaselineError(PymwalibError):
    """Raised when call to C mwalib_correlator_context_read_by_baseline fails"""
    pass

class PymwalibCorrelatorContextReadByFrequencyError(PymwalibError):
    """Raised when call to C mwalib_correlator_context_read_by_frequency fails"""
    pass

class PymwalibNoDataForTimestepAndCoarseChannelError(PymwalibError):
    """Raised when call to C mwalib functions that read data ask for data at a timestep/coarse channel where
    there is no data"""
    pass


from pymwalib.correlator_context import CorrelatorContext
import pytest
import numpy as np
from os.path import join as path_join, dirname

def prefix_test_data(path):
    return path_join(dirname(__file__), "data", path)

@pytest.fixture
def mwax_corr_context() -> CorrelatorContext:
    return CorrelatorContext(
        prefix_test_data("1297526432_mwax/1297526432.metafits"),
        list(map(
            prefix_test_data,
            [
                "1297526432_mwax/1297526432_20210216160014_ch117_000.fits",
                "1297526432_mwax/1297526432_20210216160014_ch117_001.fits",
                "1297526432_mwax/1297526432_20210216160014_ch118_000.fits",
                "1297526432_mwax/1297526432_20210216160014_ch118_001.fits"
            ]
        ))
    )

def test_mwax_antennas(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.metafits_context.antennas) == 2
    assert mwax_corr_context.metafits_context.antennas[0].tile_id == 51
    assert mwax_corr_context.metafits_context.antennas[1].tile_id == 52

def test_baselines(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.metafits_context.baselines) == 3
    assert mwax_corr_context.metafits_context.baselines[0].ant1_index == 0
    assert mwax_corr_context.metafits_context.baselines[0].ant2_index == 0
    assert mwax_corr_context.metafits_context.baselines[1].ant1_index == 0
    assert mwax_corr_context.metafits_context.baselines[1].ant2_index == 1
    assert mwax_corr_context.metafits_context.baselines[2].ant1_index == 1
    assert mwax_corr_context.metafits_context.baselines[2].ant2_index == 1

def test_mwax_rfinputs(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.metafits_context.rf_inputs) == 4
    assert mwax_corr_context.metafits_context.rf_inputs[0].tile_id == 51
    assert mwax_corr_context.metafits_context.rf_inputs[0].pol == "X"
    assert mwax_corr_context.metafits_context.rf_inputs[1].tile_id == 51
    assert mwax_corr_context.metafits_context.rf_inputs[1].pol == "Y"
    assert mwax_corr_context.metafits_context.rf_inputs[2].tile_id == 52
    assert mwax_corr_context.metafits_context.rf_inputs[2].pol == "X"
    assert mwax_corr_context.metafits_context.rf_inputs[3].tile_id == 52
    assert mwax_corr_context.metafits_context.rf_inputs[3].pol == "Y"

def test_mwax_coarse_channels(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.coarse_channels) == 2
    assert mwax_corr_context.coarse_channels[0].rec_chan_number == 117
    assert mwax_corr_context.coarse_channels[1].rec_chan_number == 118

    #
    # NOTE due to this being a modified metafits, the below is true.
    # With the "real" metafits the corr_chan_numbers would be 8 and 9 respectively
    #
    assert mwax_corr_context.coarse_channels[0].corr_chan_number == 0
    assert mwax_corr_context.coarse_channels[1].corr_chan_number == 1

def test_mwax_timesteps(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.timesteps) == 592
    assert mwax_corr_context.timesteps[0].unix_time_ms == 1613491214000
    assert mwax_corr_context.timesteps[1].unix_time_ms == 1613491214500
    assert mwax_corr_context.timesteps[590].unix_time_ms == 1613491509000
    assert mwax_corr_context.timesteps[591].unix_time_ms == 1613491509500

def test_read_by_baseline(mwax_corr_context: CorrelatorContext):
    ts = 0
    chan = 0
    data_by_bl = mwax_corr_context.read_by_baseline(ts, chan)
    data_by_f = mwax_corr_context.read_by_frequency(ts, chan)

    # Test length
    assert len(data_by_bl) == mwax_corr_context.num_timestep_coarse_chan_floats
    assert len(data_by_f) == mwax_corr_context.num_timestep_coarse_chan_floats

    # Sum them and compare
    sum_bl = np.sum(data_by_bl)
    sum_f = np.sum(data_by_f)
    assert sum_bl == sum_f
    print(f"\nCorrelator Sum by baseline  == {sum_bl}")
    print(f"Correlator Sum by frequency == {sum_f}")

def test_common_timestep_indices(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.common_timestep_indices) == mwax_corr_context.num_common_timesteps
    assert len(mwax_corr_context.common_timestep_indices) == 2
    assert mwax_corr_context.common_timestep_indices[0] == 0
    assert mwax_corr_context.common_timestep_indices[1] == 1

def test_common_good_timestep_indices(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.common_good_timestep_indices) == mwax_corr_context.num_common_good_timesteps
    assert len(mwax_corr_context.common_good_timestep_indices) == 1
    assert mwax_corr_context.common_good_timestep_indices[0] == 1

def test_provided_timestep_indices(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.provided_timestep_indices) == mwax_corr_context.num_provided_timesteps
    assert len(mwax_corr_context.provided_timestep_indices) == 4
    assert mwax_corr_context.provided_timestep_indices[0] == 0
    assert mwax_corr_context.provided_timestep_indices[1] == 1
    assert mwax_corr_context.provided_timestep_indices[2] == 160
    assert mwax_corr_context.provided_timestep_indices[3] == 161

def test_common_coarse_chan_indices(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.common_coarse_chan_indices) == mwax_corr_context.num_common_coarse_chans
    assert len(mwax_corr_context.common_coarse_chan_indices) == 2
    assert mwax_corr_context.common_coarse_chan_indices[0] == 0
    assert mwax_corr_context.common_coarse_chan_indices[1] == 1

def test_common_good_coarse_chan_indices(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.common_good_coarse_chan_indices) == mwax_corr_context.num_common_good_coarse_chans
    assert len(mwax_corr_context.common_good_coarse_chan_indices) == 2
    assert mwax_corr_context.common_good_coarse_chan_indices[0] == 0
    assert mwax_corr_context.common_good_coarse_chan_indices[1] == 1

def test_provided_coarse_chan_indices(mwax_corr_context: CorrelatorContext):
    assert len(mwax_corr_context.provided_coarse_chan_indices) == mwax_corr_context.num_provided_coarse_chans
    assert len(mwax_corr_context.provided_coarse_chan_indices) == 2
    assert mwax_corr_context.provided_coarse_chan_indices[0] == 0
    assert mwax_corr_context.provided_coarse_chan_indices[1] == 1



#!/usr/bin/env python
#
# VoltageContext: main interface for pymwalib for voltage data
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import numpy.ctypeslib as npct
from .mwalib import CVoltageContextS,ct,mwalib_library,create_string_buffer,CVoltageMetadataS,MWALIB_SUCCESS,MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN
from .common import ERROR_MESSAGE_LEN,MWAVersion
from .errors import PymwalibVoltageMetadataGetError,PymwalibVoltageContextNewError,PymwalibCorrelatorContextDisplayError,PymwalibNoDataForTimestepAndCoarseChannelError,PymwalibVoltageContextReadFileError,PymwalibVoltageContextReadSecondError
from .coarse_channel import CoarseChannel
from .metafits_metadata import MetafitsMetadata
from .timestep import TimeStep
from .version import check_mwalib_version


class VoltageContext:
    """Main class to interface with mwalib"""
    def __init__(self, metafits_filename: str, voltage_filenames: list):
        """Take metafits and voltage files, and populate this class via mwalib"""
        #
        # Ensure we have a compatible version of mwalib
        #
        check_mwalib_version()

        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)
        self._voltage_context_object = ct.POINTER(CVoltageContextS)()

        # First populate the context object
        self._get_voltage_context(metafits_filename, voltage_filenames)

        # Now Get voltage metadata
        c_object_ptr = ct.POINTER(CVoltageMetadataS)()
        if mwalib_library.mwalib_voltage_metadata_get(self._voltage_context_object,
                                                      ct.byref(c_object_ptr),
                                                      error_message,
                                                      ERROR_MESSAGE_LEN) != 0:
            raise PymwalibVoltageMetadataGetError(
                f"Error creating voltage metadata object: {error_message.decode('utf-8').rstrip()}")

        c_object = c_object_ptr.contents

        # Populate all the fields
        self.mwa_version: MWAVersion = MWAVersion(c_object.mwa_version)

        self.num_timesteps = c_object.num_timesteps
        self.timestep_duration_ms: int = c_object.timestep_duration_ms
        self.num_coarse_chans = c_object.num_coarse_chans

        self.num_common_timesteps = c_object.num_common_timesteps
        self.num_common_coarse_chans = c_object.num_common_coarse_chans
        self.common_start_unix_time_ms = c_object.common_start_unix_time_ms
        self.common_end_unix_time_ms = c_object.common_end_unix_time_ms
        self.common_start_gps_time_ms = c_object.common_start_gps_time_ms
        self.common_end_gps_time_ms = c_object.common_end_gps_time_ms
        self.common_duration_ms = c_object.common_duration_ms
        self.common_bandwidth_hz = c_object.common_bandwidth_hz

        self.num_common_good_timesteps = c_object.num_common_good_timesteps
        self.num_common_good_coarse_chans = c_object.num_common_good_coarse_chans
        self.common_good_start_unix_time_ms = c_object.common_good_start_unix_time_ms
        self.common_good_end_unix_time_ms = c_object.common_good_end_unix_time_ms
        self.common_good_start_gps_time_ms = c_object.common_good_start_gps_time_ms
        self.common_good_end_gps_time_ms = c_object.common_good_end_gps_time_ms
        self.common_good_duration_ms = c_object.common_good_duration_ms
        self.common_good_bandwidth_hz = c_object.common_good_bandwidth_hz

        self.num_provided_timesteps = c_object.num_provided_timesteps
        self.num_provided_coarse_chans = c_object.num_provided_coarse_chans

        self.coarse_chan_width_hz: int = c_object.coarse_chan_width_hz
        self.fine_chan_width_hz: int = c_object.fine_chan_width_hz
        self.num_fine_chans_per_coarse: int = c_object.num_fine_chans_per_coarse

        self.sample_size_bytes: int = c_object.sample_size_bytes
        self.num_voltage_blocks_per_timestep: int = c_object.num_voltage_blocks_per_timestep
        self.num_voltage_blocks_per_second: int = c_object.num_voltage_blocks_per_second
        self.num_samples_per_voltage_block: int = c_object.num_samples_per_voltage_block
        self.voltage_block_size_bytes: int = c_object.voltage_block_size_bytes
        self.delay_block_size_bytes: int = c_object.delay_block_size_bytes
        self.data_file_header_size_bytes: int = c_object.data_file_header_size_bytes
        self.expected_voltage_data_file_size_bytes: int = c_object.expected_voltage_data_file_size_bytes

        # Common timesteps
        self.common_timestep_indices = []
        for i in range(self.num_common_timesteps):
            self.common_timestep_indices.append(c_object.common_timestep_indices[i])

        # Common coarse chans
        self.common_coarse_chans = []
        for i in range(self.num_common_coarse_chans):
            self.common_coarse_chans.append(c_object.common_coarse_chan_indices[i])

        # Common Good timesteps
        self.common_good_timestep_indices = []
        for i in range(self.num_common_good_timesteps):
            self.common_good_timestep_indices.append(c_object.common_good_timestep_indices[i])

        # Common Good coarse chans
        self.common_good_coarse_chans = []
        for i in range(self.num_common_good_coarse_chans):
            self.common_good_coarse_chans.append(c_object.common_good_coarse_chan_indices[i])

        # Provided timesteps
        self.provided_timestep_indices = []
        for i in range(self.num_provided_timesteps):
            self.provided_timestep_indices.append(c_object.provided_timestep_indices[i])

        # Provided coarse chans
        self.provided_coarse_chan_indices = []
        for i in range(self.num_provided_coarse_chans):
            self.provided_coarse_chan_indices.append(c_object.provided_coarse_chan_indices[i])

        # Now Get metafits metadata
        self.metafits_context: MetafitsMetadata = MetafitsMetadata(None, None, self._voltage_context_object)

        # Populate coarse channels
        self.coarse_channels = CoarseChannel.get_correlator_or_voltage_coarse_channels(c_object_ptr.contents)

        # Populate timesteps
        self.timesteps = TimeStep.get_correlator_or_voltage_timesteps(c_object_ptr.contents)

        # We're now finished with the C memory, so free it
        mwalib_library.mwalib_voltage_metadata_free(c_object)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure we free the rust allocated memory in mwalib"""
        if self._voltage_context_object:
            mwalib_library.mwalib_voltage_context_free(self._voltage_context_object)

    def _get_voltage_context(self, metafits_filename: str, voltage_filenames: list):
        """This method will read and validate the metafits and voltage files. If all has worked, then
        the context object can be used in subsequent calls to populate aspects of this class."""
        if mwalib_library:
            # Encode all inputs as UTF-8.
            m = ct.c_char_p(metafits_filename.encode("utf-8"))

            # https://stackoverflow.com/questions/4145775/how-do-i-convert-a-python-list-into-a-c-array-by-using-ctypes
            encoded = []
            for v in voltage_filenames:
                encoded.append(ct.c_char_p(v.encode("utf-8")))
            seq = ct.c_char_p * len(encoded)
            g = seq(*encoded)
            error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

            if mwalib_library.mwalib_voltage_context_new(
                m, g, len(encoded), ct.byref(self._voltage_context_object), error_message, ERROR_MESSAGE_LEN) != 0:
                raise PymwalibVoltageContextNewError(f"Error creating voltage context object: "
                                             f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise PymwalibVoltageContextNewError(f"Error creating voltage context object: mwalib is not loaded.")

    def display(self):
         """Displays a human readable summary of the voltage context"""
         error_message = create_string_buffer(ERROR_MESSAGE_LEN)

         if mwalib_library.mwalib_voltage_context_display(self._voltage_context_object,
                                                  error_message,
                                                  ERROR_MESSAGE_LEN) != 0:
             raise PymwalibCorrelatorContextDisplayError(f"Error calling mwalib_voltage_context_display(): "
                                                        f"{error_message.decode('utf-8').rstrip()}")

    def read_file(self, timestep_index: int, coarse_chan_index: int):
         """Retrieve one file of VCS data as a numpy array."""
         error_message = " ".encode("utf-8") * ERROR_MESSAGE_LEN

         byte_buffer_len = self.voltage_block_size_bytes * \
                           self.num_voltage_blocks_per_timestep

         byte_buffer_type = ct.c_byte * byte_buffer_len
         buffer = byte_buffer_type()

         ret_val = mwalib_library.mwalib_voltage_context_read_file(self._voltage_context_object,
                                                    ct.c_size_t(timestep_index),
                                                    ct.c_size_t(coarse_chan_index),
                                                    buffer,
                                                    byte_buffer_len,
                                                    error_message, ERROR_MESSAGE_LEN)

         if ret_val == MWALIB_SUCCESS:
             return npct.as_array(buffer, shape=(byte_buffer_len,))
         elif ret_val == MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN:
             raise PymwalibNoDataForTimestepAndCoarseChannelError(f"No data exists for this timestep {timestep_index} and coarse channel {coarse_chan_index}")
         else:
             raise PymwalibVoltageContextReadFileError(f"Error reading data: "
                                                         f"{error_message.decode('utf-8').rstrip()}")


    def read_second(self, gps_second_start: int , gps_second_count: int, coarse_chan_index: int):
         """Retrieve multiple seconds of VCS data as a numpy array."""
         error_message = " ".encode("utf-8") * ERROR_MESSAGE_LEN

         byte_buffer_len = self.voltage_block_size_bytes * self.num_voltage_blocks_per_second * gps_second_count

         byte_buffer_type = ct.c_byte * byte_buffer_len
         buffer = byte_buffer_type()

         ret_val = mwalib_library.mwalib_voltage_context_read_second(self._voltage_context_object,
                                                      ct.c_ulong(gps_second_start),
                                                      ct.c_size_t(gps_second_count),
                                                      ct.c_size_t(coarse_chan_index),
                                                      buffer,
                                                      byte_buffer_len,
                                                      error_message, ERROR_MESSAGE_LEN)

         if ret_val == MWALIB_SUCCESS:
             return npct.as_array(buffer, shape=(byte_buffer_len,))
         elif ret_val == MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN:
             raise PymwalibNoDataForTimestepAndCoarseChannelError(f"Not all data exists for {gps_second_start} (for {gps_second_count} sec) and coarse channel {coarse_chan_index}")
         else:
             raise PymwalibVoltageContextReadSecondError(f"Error reading data: "
                                                             f"{error_message.decode('utf-8').rstrip()}")

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(\n" \
               f"MWA Version                           : {MWAVersion(self.mwa_version).name}\n" \
               f"Num timesteps                         : {self.num_timesteps}\n" \
               f"Tmestep duration (ms)                 : {self.timestep_duration_ms} ms\n" \
               f"Num coarse channels                   : {self.num_coarse_chans}\n" \
               f"num_provided_timestep_indices         : {self.num_provided_timesteps}\n" \
               f"num_provided_coarse_chan_indices      : {self.num_provided_coarse_chans}\n" \
               f"(common) Start time (UNIX)            : {float(self.common_start_unix_time_ms) / 1000.}\n" \
               f"(common) End time (UNIX)              : {float(self.common_end_unix_time_ms) / 1000.}\n" \
               f"(common) Start time (GPS)             : {float(self.common_start_gps_time_ms) / 1000.}\n" \
               f"(common) End time (GPS)               : {float(self.common_end_gps_time_ms) / 1000.}\n" \
               f"(common) Duration                     : {float(self.common_duration_ms) / 1000.} s\n" \
               f"(common) num timesteps                : {self.num_common_timesteps}\n" \
               f"(common) num coarse channels          : {self.num_common_coarse_chans}\n" \
               f"(common) Bandwidth                    : {float(self.common_bandwidth_hz) / 1000000.} MHz\n" \
               f"(common good) Start time (UNIX)       : {float(self.common_good_start_unix_time_ms) / 1000.}\n" \
               f"(common good) End time (UNIX)         : {float(self.common_good_end_unix_time_ms) / 1000.}\n" \
               f"(common good) Start time (GPS)        : {float(self.common_good_start_gps_time_ms) / 1000.}\n" \
               f"(common good) End time (GPS)          : {float(self.common_good_end_gps_time_ms) / 1000.}\n" \
               f"(common good) Duration                : {float(self.common_good_duration_ms) / 1000.} s\n" \
               f"(common good) num timesteps           : {self.num_common_good_timesteps}\n" \
               f"(common good) num coarse channels     : {self.num_common_good_coarse_chans}\n" \
               f"(common good) Bandwidth               : {float(self.common_good_bandwidth_hz) / 1000000.} MHz\n" \
               f"Coarse channel width                  : {float(self.coarse_chan_width_hz) / 1000000.} MHz\n" \
               f"Fine channel width                    : {float(self.fine_chan_width_hz) / 1000.} kHz\n" \
               f"Num fine channels per coarse          : {self.num_fine_chans_per_coarse}\n" \
               f"Sample size                           : {self.sample_size_bytes} bytes\n" \
               f"Num voltage blocks per timeste        : {self.num_voltage_blocks_per_timestep}\n" \
               f"Num voltage blocks per second         : {self.num_voltage_blocks_per_second}\n" \
               f"Num samples per voltage block         : {self.num_samples_per_voltage_block}\n" \
               f"Voltage block size                    : {self.voltage_block_size_bytes} bytes\n" \
               f"Delay block size                      : {self.delay_block_size_bytes} bytes\n" \
               f"Data file header size                 : {self.data_file_header_size_bytes} bytes\n" \
               f"Expected voltage data file size       : {self.expected_voltage_data_file_size_bytes} bytes\n)\n"



#!/usr/bin/env python
#
# Common classes and functions
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import enum

ERROR_MESSAGE_LEN = 1024

class MWAVersion(enum.Enum):
    """Enum for MWA version"""
    # MWA correlator(v1.0), having data files without any batch numbers.
    CorrOldLegacy = 1
    # MWA correlator(v1.0), having data files with "gpubox" and batch numbers in their names.
    CorrLegacy = 2
    # MWAX correlator(v2.0)
    CorrMWAXv2 = 3
    # Legacy VCS Recombined
    VCSLegacyRecombined = 4
    # MWAX VCS
    VCSMWAXv2 = 5

class VisPol(enum.Enum):
    """Enum representing the four visibility polarisations for MWA"""
    XX = 1
    XY = 2
    YX = 3
    YY = 4

class GeometricDelaysApplied(enum.Enum):
    """Enum representing the state of Geometric delays applied for this observation"""
    No = 0
    Zenith = 1
    TilePointing = 2
    AzElTracking = 3

class MWAMode(enum.Enum):
    """Enum representing the correlator mode"""
    No_Capture = 0
    Burst_Vsib = 1
    Sw_Cor_Vsib = 2
    Hw_Cor_Pkts = 3
    Rts_32t = 4
    Hw_Lfiles = 5
    Hw_Lfiles_Nomentok = 6
    Sw_Cor_Vsib_Nomentok = 7
    Burst_Vsib_Synced = 8
    Burst_Vsib_Raw = 9
    Lfiles_Client = 16
    No_Capture_Burst = 17
    Enter_Burst = 18
    Enter_Channel = 19
    Voltage_Raw = 20
    Corr_Mode_Change = 21
    Voltage_Start = 22
    Voltage_Stop = 23
    Voltage_Buffer = 24
    Mwax_Correlator = 30
    Mwax_Vcs = 31



#!/usr/bin/env python
#
# baseline: class representing a single baseline in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from .mwalib import CBaselineS, CMetafitsMetadataS


class Baseline:
    """
    A class representing a single Baseline

    Attributes
    ----------
    index : int
        Ordinal index of this time step.

    ant1_index : int
        The index in the antenna array for the first member of the baseline

    ant2_index : int
        The index in the antenna array for the second member of the baseline

    """

    def __init__(self,
                 index: int,
                 ant1_index: int,
                 ant2_index: int):
        """Initialise the class"""
        self.index: int = index
        self.ant1_index: int = ant1_index
        self.ant2_index: int = ant2_index

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Order: {self.index}, " \
               f"Antennas: {self.ant1_index} v {self.ant2_index})"

    @staticmethod
    def get_baselines(metafits_metadata: CMetafitsMetadataS) -> []:
        """Retrieve all of the baseline metadata and populate a list of baselines."""
        baselines = []

        for i in range(0, metafits_metadata.num_baselines):
            obj: CBaselineS = metafits_metadata.baselines[i]

            # Populate all the fields
            baselines.append(Baseline(i,
                                      obj.ant1_index,
                                      obj.ant2_index))

        return baselines


#!/usr/bin/env python
#
# mwalib.py: C wrappers
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import ctypes as ct
import sys

MWALIB_SUCCESS = 0
MWALIB_FAILURE = 1
MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN = -1

#
# Creates a string buffer for interacting with C strings
#
def create_string_buffer(length: int) -> bytes:
    return " ".encode("utf-8") * length


#
# C MetafitsContext struct
#
class CMetafitsContextS(ct.Structure):
    pass


#
# C CorrelatorContext struct
#
class CCorrelatorContextS(ct.Structure):
    pass


#
# C VoltageContext struct
#
class CVoltageContextS(ct.Structure):
    pass


#
# mwalib: setup linking to the mwalib library
#
prefix = {"win32": ""}.get(sys.platform, "lib")
extension = {"darwin": ".dylib", "win32": ".dll"}.get(sys.platform, ".so")
mwalib_filename = prefix + "mwalib" + extension
mwalib_library = None

try:
    mwalib_library = ct.cdll.LoadLibrary(mwalib_filename)
except Exception as library_load_err:
    print(f"Error loading {mwalib_filename}. Please check that it is in your system library path or in your LD_LIBRARY_PATH "
          f"environment variable.\n\nError was: {library_load_err}")
    exit(1)

#
# Define the Library functions if the library was loaded
#
if mwalib_library:
    #
    # mwalib_get_version_major
    #
    mwalib_library.mwalib_get_version_major.argtypes = None
    mwalib_library.mwalib_get_version_major.restype = ct.c_uint32

    #
    # mwalib_get_version_minor
    #
    mwalib_library.mwalib_get_version_minor.argtypes = None
    mwalib_library.mwalib_get_version_minor.restype = ct.c_uint32

    #
    # mwalib_get_version_patch
    #
    mwalib_library.mwalib_get_version_patch.argtypes = None
    mwalib_library.mwalib_get_version_patch.restype = ct.c_uint32

    #
    # mwalib_metafits_context_new()
    #
    mwalib_library.mwalib_metafits_context_new.argtypes = \
        (ct.c_char_p,                               # metafits
         ct.c_uint,                                 # MWAVersion
         ct.POINTER(ct.POINTER(CMetafitsContextS)), # Pointer to pointer to CorrelatorContext
         ct.c_char_p,                               # error message
         ct.c_size_t)                               # length of error message

    #
    # mwalib_metafits_context_free()
    #
    mwalib_library.mwalib_metafits_context_free.argtypes = (ct.POINTER(CMetafitsContextS),)
    mwalib_library.mwalib_metafits_context_free.restype = ct.c_int32

    #
    # mwalib_metafits_context_display()
    #
    mwalib_library.mwalib_metafits_context_display.argtypes = (ct.POINTER(CMetafitsContextS),)
    mwalib_library.mwalib_metafits_context_display.restype = ct.c_int32


    #
    # C Antenna struct
    #
    class CAntennaS(ct.Structure):
        _fields_ = [('ant', ct.c_uint32),
                    ('tile_id', ct.c_uint32),
                    ('tile_name', ct.c_char_p),
                    ('rfinput_x', ct.c_size_t),
                    ('rfinput_y', ct.c_size_t),
                    ('electrical_length_m', ct.c_double),
                    ('north_m', ct.c_double),
                    ('east_m', ct.c_double),
                    ('height_m', ct.c_double),
                    ]


    #
    # C Baseline struct
    #
    class CBaselineS(ct.Structure):
        _fields_ = [('ant1_index', ct.c_size_t),
                    ('ant2_index', ct.c_size_t), ]


    #
    # C CoarseChannel struct
    #
    class CCoarseChannelS(ct.Structure):
        _fields_ = [('corr_chan_number', ct.c_size_t),
                    ('rec_chan_number', ct.c_size_t),
                    ('gpubox_number', ct.c_size_t),
                    ('chan_width_hz', ct.c_uint32),
                    ('chan_start_hz', ct.c_uint32),
                    ('chan_centre_hz', ct.c_uint32),
                    ('chan_end_hz', ct.c_uint32), ]


    #
    # C RFInput struct
    #
    class CRFInputS(ct.Structure):
        _fields_ = [('input', ct.c_uint32),
                    ('ant', ct.c_uint32),
                    ('tile_id', ct.c_uint32),
                    ('tile_name', ct.c_char_p),
                    ('pol', ct.c_char_p),
                    ('electrical_length_m', ct.c_double),
                    ('north_m', ct.c_double),
                    ('east_m', ct.c_double),
                    ('height_m', ct.c_double),
                    ('vcs_order', ct.c_uint32),
                    ('subfile_order', ct.c_uint32),
                    ('flagged', ct.c_bool),
                    ('rec_number', ct.c_uint32),
                    ('rec_slot_number', ct.c_uint32)]


    #
    # C TimeStep struct
    #
    class CTimeStepS(ct.Structure):
        _fields_ = [('unix_time_ms', ct.c_uint64),
                    ('gps_time_ms', ct.c_uint64), ]


    #
    # C MetafitsMetadata struct
    #
    class CMetafitsMetadataS(ct.Structure):
        _fields_ = [('obs_id', ct.c_uint32),
                    ('global_analogue_attenuation_db', ct.c_double),
                    ('ra_tile_pointing_deg', ct.c_double),
                    ('dec_tile_pointing_deg', ct.c_double),
                    ('ra_phase_center_deg', ct.c_double),
                    ('dec_phase_center_deg', ct.c_double),
                    ('az_deg', ct.c_double),
                    ('alt_deg', ct.c_double),
                    ('za_deg', ct.c_double),
                    ('az_rad', ct.c_double),
                    ('alt_rad', ct.c_double),
                    ('za_rad', ct.c_double),
                    ('sun_alt_deg', ct.c_double),
                    ('sun_distance_deg', ct.c_double),
                    ('moon_distance_deg', ct.c_double),
                    ('jupiter_distance_deg', ct.c_double),
                    ('lst_deg', ct.c_double),
                    ('lst_rad', ct.c_double),
                    ('hour_angle_string', ct.c_char_p),
                    ('grid_name', ct.c_char_p),
                    ('grid_number', ct.c_int32),
                    ('creator', ct.c_char_p),
                    ('project_id', ct.c_char_p),
                    ('obs_name', ct.c_char_p),
                    ('mode', ct.c_uint32),
                    ('geometric_delays_applied', ct.c_uint32),
                    ('cable_delays_applied', ct.c_bool),
                    ('calibration_delays_and_gains_applied', ct.c_bool),
                    ('corr_fine_chan_width_hz', ct.c_uint32),
                    ('corr_int_time_ms', ct.c_uint64),
                    ('num_corr_fine_chans_per_coarse', ct.c_size_t),
                    ('receivers', ct.POINTER(ct.c_size_t)),
                    ('delays', ct.POINTER(ct.c_uint32)),
                    ('sched_start_utc', ct.c_uint64),
                    ('sched_end_utc', ct.c_uint64),
                    ('sched_start_mjd', ct.c_double),
                    ('sched_end_mjd', ct.c_double),
                    ('sched_start_unix_time_ms', ct.c_uint64),
                    ('sched_end_unix_time_ms', ct.c_uint64),
                    ('sched_start_gps_time_ms', ct.c_uint64),
                    ('sched_end_gps_time_ms', ct.c_uint64),
                    ('sched_duration_ms', ct.c_uint64),
                    ('quack_time_duration_ms', ct.c_uint64),
                    ('good_time_unix_ms', ct.c_uint64),
                    ('good_time_gps_ms', ct.c_uint64),
                    ('num_ants', ct.c_size_t),
                    ('antennas', ct.POINTER(CAntennaS)),
                    ('num_rf_inputs', ct.c_size_t),
                    ('rf_inputs', ct.POINTER(CRFInputS)),
                    ('num_ant_pols', ct.c_size_t),
                    ('num_baselines', ct.c_size_t),
                    ('baselines', ct.POINTER(CBaselineS)),
                    ('num_visibility_pols', ct.c_size_t),
                    ('num_metafits_coarse_chans', ct.c_size_t),
                    ('metafits_coarse_chans', ct.POINTER(CCoarseChannelS)),
                    ('num_metafits_timesteps', ct.c_size_t),
                    ('metafits_timesteps', ct.POINTER(CTimeStepS)),
                    ('obs_bandwidth_hz', ct.c_uint32),
                    ('coarse_chan_width_hz', ct.c_uint32),
                    ('centre_freq_hz', ct.c_uint32),
                    ('metafits_filename', ct.c_char_p)
                    ]

    #
    # mwalib_metafits_metadata_get()
    #
    mwalib_library.mwalib_metafits_metadata_get.argtypes = \
        (ct.POINTER(CMetafitsContextS),                 # metafits context pointer OR
         ct.POINTER(CCorrelatorContextS),               # correlator context pointer OR
         ct.POINTER(CVoltageContextS),                  # voltage context pointer
         ct.POINTER(ct.POINTER(CMetafitsMetadataS)),    # Pointer to pointer to CMetafitsMetadataS
         ct.c_char_p,                                   # error message
         ct.c_size_t)                                   # length of error message
    mwalib_library.mwalib_metafits_metadata_get.restype = ct.c_int32

    #
    # mwalib_metafits_metadata_free()
    #
    mwalib_library.mwalib_metafits_metadata_free.argtypes = (ct.POINTER(CMetafitsMetadataS),)
    mwalib_library.mwalib_metafits_metadata_free.restype = ct.c_int32

    #
    # mwalib_correlator_context_new()
    #
    mwalib_library.mwalib_correlator_context_new.argtypes = \
        (ct.c_char_p,                                   # metafits
         ct.POINTER(ct.c_char_p),                       # gpuboxes files array
         ct.c_size_t,                                   # gpubox count
         ct.POINTER(ct.POINTER(CCorrelatorContextS)),   # Pointer to pointer to CorrelatorContext
         ct.c_char_p,                                   # error message
         ct.c_size_t)                                   # length of error message
    mwalib_library.mwalib_correlator_context_new.restype = ct.c_int32

    #
    # mwalib_correlator_context_free()
    #
    mwalib_library.mwalib_correlator_context_free.argtypes = (ct.POINTER(CCorrelatorContextS),)
    mwalib_library.mwalib_correlator_context_free.restype = ct.c_int32

    #
    # mwalib_correlator_context_display()
    #
    mwalib_library.mwalib_correlator_context_display.argtypes = (ct.POINTER(CCorrelatorContextS),)
    mwalib_library.mwalib_correlator_context_display.restype = ct.c_int32

    #
    # mwalib_correlator_context_read_by_baseline()
    #
    mwalib_library.mwalib_correlator_context_read_by_baseline.argtypes = \
        (ct.POINTER(CCorrelatorContextS),   # context
         ct.c_size_t,                       # input timestep_index
         ct.c_size_t,                       # input coarse_chan_index
         ct.POINTER(ct.c_float),            # buffer_ptr
         ct.c_size_t,                       # buffer_len
         ct.c_char_p,                       # error message
         ct.c_size_t)                        # length of error message
    mwalib_library.mwalib_correlator_context_read_by_baseline.restype = ct.c_int32

    #
    # mwalib_correlator_context_read_by_frequency()
    #
    mwalib_library.mwalib_correlator_context_read_by_frequency.argtypes = \
        (ct.POINTER(CCorrelatorContextS),   # context
         ct.c_size_t,                       # input timestep_index
         ct.c_size_t,                       # input coarse_chan_index
         ct.POINTER(ct.c_float),            # buffer_ptr
         ct.c_size_t,                       # buffer_len
         ct.c_char_p,                       # error message
         ct.c_size_t)                       # length of error message
    mwalib_library.mwalib_correlator_context_read_by_frequency.restype = ct.c_int32

    #
    # C CorrelatorMetadata struct
    #
    class CCorrelatorMetadataS(ct.Structure):
        _fields_ = [('mwa_version', ct.c_uint32),
                    ('timesteps', ct.POINTER(CTimeStepS)),
                    ('num_timesteps', ct.c_size_t),
                    ('coarse_chans', ct.POINTER(CCoarseChannelS)),
                    ('num_coarse_chans', ct.c_size_t),

                    ('num_common_timesteps', ct.c_size_t),
                    ('common_timestep_indices', ct.POINTER(ct.c_size_t)),
                    ('num_common_coarse_chans', ct.c_size_t),
                    ('common_coarse_chan_indices', ct.POINTER(ct.c_size_t)),
                    ('common_start_unix_time_ms', ct.c_uint64),
                    ('common_end_unix_time_ms', ct.c_uint64),
                    ('common_start_gps_time_ms', ct.c_uint64),
                    ('common_end_gps_time_ms', ct.c_uint64),
                    ('common_duration_ms', ct.c_uint64),
                    ('common_bandwidth_hz', ct.c_uint32),

                    ('num_common_good_timesteps', ct.c_size_t),
                    ('common_good_timestep_indices', ct.POINTER(ct.c_size_t)),
                    ('num_common_good_coarse_chans', ct.c_size_t),
                    ('common_good_coarse_chan_indices', ct.POINTER(ct.c_size_t)),
                    ('common_good_start_unix_time_ms', ct.c_uint64),
                    ('common_good_end_unix_time_ms', ct.c_uint64),
                    ('common_good_start_gps_time_ms', ct.c_uint64),
                    ('common_good_end_gps_time_ms', ct.c_uint64),
                    ('common_good_duration_ms', ct.c_uint64),
                    ('common_good_bandwidth_hz', ct.c_uint32),

                    ('num_provided_timesteps', ct.c_size_t),
                    ('provided_timestep_indices', ct.POINTER(ct.c_size_t)),
                    ('num_provided_coarse_chans', ct.c_size_t),
                    ('provided_coarse_chan_indices', ct.POINTER(ct.c_size_t)),

                    ('num_timestep_coarse_chan_bytes', ct.c_size_t),
                    ('num_timestep_coarse_chan_floats', ct.c_size_t),
                    ('num_gpubox_files', ct.c_size_t)
                    ]
    #
    # mwalib_correlator_metadata_get()
    #
    mwalib_library.mwalib_correlator_metadata_get.argtypes = \
        (ct.POINTER(CCorrelatorContextS),               # correlator context pointer
         ct.POINTER(ct.POINTER(CCorrelatorMetadataS)),  # Pointer to pointer to CCorrelatorMetadataS
         ct.c_char_p,                                   # error message
         ct.c_size_t)                                   # length of error message
    mwalib_library.mwalib_correlator_metadata_get.restype = ct.c_int32

    #
    # mwalib_correlator_metadata_free()
    #
    mwalib_library.mwalib_correlator_metadata_free.argtypes = (ct.POINTER(CCorrelatorMetadataS),)
    mwalib_library.mwalib_correlator_metadata_free.restype = ct.c_int32

    #
    # C VoltageMetadata struct
    #
    class CVoltageMetadataS(ct.Structure):
        _fields_ = [('mwa_version', ct.c_uint32),
                    ('timesteps', ct.POINTER(CTimeStepS)),
                    ('num_timesteps', ct.c_size_t),
                    ('timestep_duration_ms', ct.c_uint64),
                    ('coarse_chans', ct.POINTER(CCoarseChannelS)),
                    ('num_coarse_chans', ct.c_size_t),

                    ('num_common_timesteps', ct.c_size_t),
                    ('common_timestep_indices', ct.POINTER(ct.c_size_t)),
                    ('num_common_coarse_chans', ct.c_size_t),
                    ('common_coarse_chan_indices', ct.POINTER(ct.c_size_t)),
                    ('common_start_unix_time_ms', ct.c_uint64),
                    ('common_end_unix_time_ms', ct.c_uint64),
                    ('common_start_gps_time_ms', ct.c_uint64),
                    ('common_end_gps_time_ms', ct.c_uint64),
                    ('common_duration_ms', ct.c_uint64),
                    ('common_bandwidth_hz', ct.c_uint32),

                    ('num_common_good_timesteps', ct.c_size_t),
                    ('common_good_timestep_indices', ct.POINTER(ct.c_size_t)),
                    ('num_common_good_coarse_chans', ct.c_size_t),
                    ('common_good_coarse_chan_indices', ct.POINTER(ct.c_size_t)),
                    ('common_good_start_unix_time_ms', ct.c_uint64),
                    ('common_good_end_unix_time_ms', ct.c_uint64),
                    ('common_good_start_gps_time_ms', ct.c_uint64),
                    ('common_good_end_gps_time_ms', ct.c_uint64),
                    ('common_good_duration_ms', ct.c_uint64),
                    ('common_good_bandwidth_hz', ct.c_uint32),

                    ('num_provided_timesteps', ct.c_size_t),
                    ('provided_timestep_indices', ct.POINTER(ct.c_size_t)),
                    ('num_provided_coarse_chans', ct.c_size_t),
                    ('provided_coarse_chan_indices', ct.POINTER(ct.c_size_t)),

                    ('coarse_chan_width_hz', ct.c_uint32),
                    ('fine_chan_width_hz', ct.c_uint32),
                    ('num_fine_chans_per_coarse', ct.c_size_t),
                    ('sample_size_bytes', ct.c_size_t),
                    ('num_voltage_blocks_per_timestep', ct.c_size_t),
                    ('num_voltage_blocks_per_second', ct.c_size_t),
                    ('num_samples_per_voltage_block', ct.c_size_t),
                    ('voltage_block_size_bytes', ct.c_size_t),
                    ('delay_block_size_bytes', ct.c_size_t),
                    ('data_file_header_size_bytes', ct.c_size_t),
                    ('expected_voltage_data_file_size_bytes', ct.c_size_t)
                    ]

    #
    # mwalib_voltage_metadata_get()
    #
    mwalib_library.mwalib_voltage_metadata_get.argtypes = \
        (ct.POINTER(CVoltageContextS),              # voltage context pointer
         ct.POINTER(ct.POINTER(CVoltageMetadataS)), # Pointer to pointer to CVoltageMetadata
         ct.c_char_p,                               # error message
         ct.c_size_t)                               # length of error message
    mwalib_library.mwalib_voltage_metadata_get.restype = ct.c_int32

    #
    # mwalib_voltage_metadata_free()
    #
    mwalib_library.mwalib_voltage_metadata_free.argtypes = (ct.POINTER(CVoltageMetadataS),)
    mwalib_library.mwalib_voltage_metadata_free.restype = ct.c_int32

    #
    # mwalib_voltage_context_read_file()
    #
    mwalib_library.mwalib_voltage_context_read_file.argtypes = \
        (ct.POINTER(CVoltageContextS),  # context
         ct.c_size_t,  # input timestep_index
         ct.c_size_t,  # input coarse_chan_index
         ct.POINTER(ct.c_byte),  # buffer_ptr
         ct.c_size_t,  # buffer_len
         ct.c_char_p,  # error message
         ct.c_size_t)  # length of error message
    mwalib_library.mwalib_voltage_context_read_file.restype = ct.c_int32

    #
    # mwalib_voltage_context_read_second()
    #
    mwalib_library.mwalib_voltage_context_read_second.argtypes = \
        (ct.POINTER(CVoltageContextS),  # context
         ct.c_ulong,   # input gps second start
         ct.c_size_t,  # input gps second count
         ct.c_size_t,  # input coarse_chan_index
         ct.POINTER(ct.c_byte),  # buffer_ptr
         ct.c_size_t,  # buffer_len
         ct.c_char_p,  # error message
         ct.c_size_t)  # length of error message
    mwalib_library.mwalib_voltage_context_read_second.restype = ct.c_int32



#!/usr/bin/env python
#
# Constants (derived from the mwalib rust public constants)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#

#
# the velocity factor of electic fields in RG-6 like coax cable
#
MWA_COAX_V_FACTOR = 1.204

# The MWA's altitude in metres.
MWA_ALTITUDE_METRES = 377.827

# The MWA's latitude on Earth in radians. This is -26d42m11.94986s.
MWA_LATITUDE_RADIANS = -0.4660608448386394

# The MWA's longitude on Earth in radians. This is 116d40m14.93485s.
MWA_LONGITUDE_RADIANS = 2.0362898668561042


#!/usr/bin/env python
#
# pymwalib examples/sum-vcs - utilise all cores to sum the vcs data files and compare against single threaded
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# NOTE: this example requires numpy and joblib packages. These can be installed via pip.
# e.g. pip install numpy
#      pip install joblib
#
import argparse
import numpy as np
from pymwalib.voltage_context import VoltageContext
from pymwalib.version import check_mwalib_version
from pymwalib.errors import PymwalibNoDataForTimestepAndCoarseChannelError
import time


def sum_by_file(context: VoltageContext, timestep_index:int, coarse_chan_index: int) -> int:
    total_sum = 0

    start_gpstime_sec = int(context.timesteps[context.provided_timestep_indices[t]].gps_time_ms / 1000)
    gps_seconds_count = int(context.timestep_duration_ms/ 1000)
    end_gpstime_sec = start_gpstime_sec + gps_seconds_count

    print(f"...Summing {start_gpstime_sec} to {end_gpstime_sec} ({gps_seconds_count} seconds) and "
          f"chan: {coarse_chan_index}...", end="")

    try:
        data = context.read_file(timestep_index, coarse_chan_index)
        total_sum = data.sum(dtype=np.int64)
        print(f"{total_sum}")

    except PymwalibNoDataForTimestepAndCoarseChannelError:
        pass

    except Exception as e:
        print(f"Error: {e}")
        exit(-1)

    return total_sum


def sum_by_gps_second(context: VoltageContext, gps_start_sec, gps_end_sec, gps_seconds_count, coarse_chan_index: int) -> int:
    total_sum = 0

    print(f"...Summing {gps_start_sec} to {gps_end_sec} ({gps_seconds_count} seconds) and "
          f"chan: {coarse_chan_index}...", end="")
    try:
        data = context.read_second(gps_start_sec, gps_seconds_count, coarse_chan_index)
        total_sum = data.sum(dtype=np.int64)
        print(f"{total_sum}")

    except PymwalibNoDataForTimestepAndCoarseChannelError:
        pass

    except Exception as e:
        print(f"Error: {e}")
        exit(-1)

    return total_sum


if __name__ == "__main__":
    # ensure we have a compatible mwalib first
    # You can skip this if you want, but your first pymwalib call will raise an error. Best trap it here
    # and provide a nice user message
    try:
        check_mwalib_version()
    except Exception as e:
        print(e)
        exit(1)

    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--metafits", required=True,
                        help="Path to the metafits file.")
    parser.add_argument("datafiles", nargs='*',
                        help="Paths to the vcs data files.")
    args = parser.parse_args()

    context = VoltageContext(args.metafits, args.datafiles)

    # sum by file
    print(f"sum_by_file: Summing {context.num_provided_timesteps} timesteps "
          f"and {context.num_provided_coarse_chans} coarse channels...")
    total_sum_by_file = 0
    start_time = time.time()
    for t in context.provided_timestep_indices:
        for c in context.provided_coarse_chan_indices:
            total_sum_by_file += np.sum(sum_by_file(context, t, c))
    stop_time = time.time()
    print(f"Sum is: {total_sum_by_file} in {stop_time - start_time} seconds.\n")

    # sum by gps second
    start_gpstime_sec = int(context.timesteps[context.provided_timestep_indices[0]].gps_time_ms / 1000)
    end_gpstime_sec = int((
        context.timesteps[context.provided_timestep_indices[context.num_provided_timesteps - 1]].gps_time_ms +
        context.timestep_duration_ms) / 1000)
    gps_second_count = end_gpstime_sec - start_gpstime_sec

    print(f"sum_by_gps_second: Summing {context.num_provided_timesteps} timesteps "
          f"and {context.num_provided_coarse_chans} coarse channels...")
    total_sum_by_gps = 0
    start_time = time.time()
    for c in context.provided_coarse_chan_indices:
        total_sum_by_gps += sum_by_gps_second(context, start_gpstime_sec, end_gpstime_sec, gps_second_count, c)
    stop_time = time.time()
    print(f"Sum is: {total_sum_by_gps} in {stop_time - start_time} seconds.")



#!/usr/bin/env python
#
# pymwalib examples/sum-gpuboxes - utilise all cores to sum the hdus and compare against single threaded
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# NOTE: this example requires numpy and joblib packages. These can be installed via pip.
# e.g. pip install numpy
#      pip install joblib
#
import argparse
import numpy as np
import os
from joblib import Parallel, delayed
from pymwalib.correlator_context import CorrelatorContext
from pymwalib.version import check_mwalib_version
from pymwalib.errors import PymwalibNoDataForTimestepAndCoarseChannelError
import time


def sum_by_baseline_task(metafits_filename: str, gpubox_filenames: list, coarse_chan_index: int) -> float:
    chan_sum = 0.

    with CorrelatorContext(metafits_filename, gpubox_filenames) as context:
        if coarse_chan_index < context.num_coarse_chans:
            print(f"sum_by_baseline_task: Summing {context.num_timesteps} timesteps "
                  f"and coarse channel index {coarse_chan_index}...")

            for t in range(0, context.num_timesteps):
                try:
                    data = context.read_by_frequency(t, coarse_chan_index)
                    chan_sum += np.sum(data, dtype=np.float64)

                except PymwalibNoDataForTimestepAndCoarseChannelError:
                    pass

                except Exception as e:
                    print(f"Error: {e}")
                    exit(-1)

    return chan_sum


def sum_by_baseline_slow(metafits_filename: str, gpubox_filenames: list) -> float:
    total_sum = 0.

    with CorrelatorContext(metafits_filename, gpubox_filenames) as context:
        for coarse_chan_index in range(0, context.num_coarse_chans):
            print(f"sum_by_baseline_slow: Summing {context.num_timesteps} timesteps "
                  f"and coarse channel index {coarse_chan_index}...")
            for timestep_index in range(0, context.num_timesteps):
                try:
                    data = context.read_by_baseline(timestep_index,
                                                    coarse_chan_index)
                    total_sum += np.sum(data, dtype=np.float64)

                except PymwalibNoDataForTimestepAndCoarseChannelError:
                    pass

                except Exception as e:
                    print(f"Error: {e}")
                    exit(-1)

    return total_sum


if __name__ == "__main__":
    # ensure we have a compatible mwalib first
    # You can skip this if you want, but your first pymwalib call will raise an error. Best trap it here
    # and provide a nice user message
    try:
        check_mwalib_version()
    except Exception as e:
        print(e)
        exit(1)

    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--metafits", required=True,
                        help="Path to the metafits file.")
    parser.add_argument("gpuboxes", nargs='*',
                        help="Paths to the gpubox files.")
    args = parser.parse_args()

    # fast sum using all cores
    num_cores = len(os.sched_getaffinity(0))
    print(f"Using {num_cores} cores to fast sum all hdus...")

    start_time_fast = time.time()
    processed_list = Parallel(n_jobs=num_cores)(delayed(sum_by_baseline_task)(args.metafits, args.gpuboxes, c) for c in range(24))
    fast_sum = np.sum(processed_list)
    stop_time_fast = time.time()
    print(f"Sum is: {fast_sum} in {stop_time_fast - start_time_fast} seconds.\n")

    # slow sum restricted to one python process
    start_time_slow = time.time()
    slow_sum = sum_by_baseline_slow(args.metafits, args.gpuboxes)
    stop_time_slow = time.time()
    print(f"Sum is: {slow_sum} in {stop_time_slow - start_time_slow} seconds.")





#!/usr/bin/env python
#
# MetafitsContext: main interface for pymwalib for just viewing metafits data (no data files)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import ctypes as ct
from .mwalib import CMetafitsContextS,mwalib_library,create_string_buffer
from .common import ERROR_MESSAGE_LEN, MWAVersion
from .errors import PymwalibMetafitsContextNewError,PymwalibMetafitsContextDisplayError
from .metafits_metadata import MetafitsMetadata
from .version import check_mwalib_version


class MetafitsContext(MetafitsMetadata):
    """Main class to interface with mwalib metafits infomation"""
    def __init__(self, metafits_filename: str, mwa_version: MWAVersion):
        """Take metafits and populate this class via mwalib"""
        #
        # Ensure we have a compatible version of mwalib
        #
        check_mwalib_version()

        self._metafits_context_object = ct.POINTER(CMetafitsContextS)()

        # First populate the context object
        self._get_metafits_context(metafits_filename, mwa_version)

        # Get metafits metadata
        super().__init__(self._metafits_context_object, None, None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure we free the rust allocated memory in mwalib"""
        if self._metafits_context_object:
            mwalib_library.mwalib_metafits_context_free(self._metafits_context_object)

    def _get_metafits_context(self, metafits_filename: str, mwa_version: MWAVersion):
        """This method will read and validate the metafits and gpubox files. If all has worked, then
        the context object can be used in subsequent calls to populate aspects of this class."""
        if mwalib_library:
            # Encode all inputs as UTF-8.
            m = ct.c_char_p(metafits_filename.encode("utf-8"))

            error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)
            if mwalib_library.mwalib_metafits_context_new(
                m, mwa_version.value, ct.byref(self._metafits_context_object), error_message, ERROR_MESSAGE_LEN) != 0:
                raise PymwalibMetafitsContextNewError(f"Error creating metafits context object: "
                                             f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise PymwalibMetafitsContextNewError(f"Error creating metafits context object: mwalib.so is not loaded.")


    def display(self):
         """Displays a human readable summary of the metafits context"""
         error_message = create_string_buffer(ERROR_MESSAGE_LEN)

         if mwalib_library.mwalib_metafits_context_display(self._metafits_context_object,
                                                   error_message,
                                                   ERROR_MESSAGE_LEN) != 0:
             raise PymwalibMetafitsContextDisplayError(f"Error calling mwalib_metafits_context_display(): "
                                                        f"{error_message.decode('utf-8').rstrip()}")

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(\n" \
               f"{super().__repr__()}\n)\n"



#!/usr/bin/env python
#
# rf_input: class representing a single rf_input in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from .mwalib import CRFInputS, CMetafitsMetadataS


class RFInput:
    """
    A class representing a single RFInput

    Attributes
    ----------
    index : int
        Ordinal index of this RF Input.

    Please see https://github.com/MWATelescope/mwalib/blob/master/src/rf_input.rs for remaining attributes

    """

    def __init__(self,
                 index: int,
                 input: int,
                 ant: int,
                 tile_id: int,
                 tile_name: str,
                 pol: str,
                 electrical_length_m: float,
                 north_m: float,
                 east_m: float,
                 height_m: float,
                 vcs_order: int,
                 subfile_order: int,
                 flagged: bool,
                 rec_number: int,
                 rec_slot_number: int):
        """Initialise the class"""
        self.index: int = index
        self.input: int = input
        self.ant: int = ant
        self.tile_id: int = tile_id
        self.tile_name: str = tile_name
        self.pol: str = pol
        self.electrical_length_m: float = electrical_length_m
        self.north_m: float = north_m
        self.east_m: float = east_m
        self.height_m: float = height_m
        self.vcs_order: int = vcs_order
        self.subfile_order: int = subfile_order
        self.flagged: bool = flagged
        self.rec_number = rec_number
        self.rec_slot_number = rec_slot_number

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Index: {self.index}, " \
               f"Input: {self.input}, " \
               f"Ant: {self.ant}, " \
               f"Tile Id: {self.tile_id}, " \
               f"Tile Name: {self.tile_name}, " \
               f"Pol: {self.pol}, " \
               f"Electrical length: {self.electrical_length_m}, " \
               f"north_m: {self.north_m}, " \
               f"east_m: {self.east_m}, " \
               f"height_m: {self.height_m}, " \
               f"vcs_order: {self.vcs_order}, " \
               f"subfile_order: {self.subfile_order}, " \
               f"flagged: {self.flagged}, " \
               f"rec_number: {self.rec_number}, " \
               f"rec_slot_number: {self.rec_slot_number})"

    @staticmethod
    def get_rf_inputs(metafits_metadata: CMetafitsMetadataS) -> []:
        """Retrieve all of the rf_input metadata and populate a list of rf_inputs."""
        rf_inputs = []

        for i in range(0, metafits_metadata.num_rf_inputs):
            obj: CRFInputS = metafits_metadata.rf_inputs[i]

            # Populate all the fields
            rf_inputs.append(RFInput(i,
                                     obj.input,
                                     obj.ant,
                                     obj.tile_id,
                                     obj.tile_name.decode("utf-8"),
                                     obj.pol.decode("utf-8"),
                                     obj.electrical_length_m,
                                     obj.north_m,
                                     obj.east_m,
                                     obj.height_m,
                                     obj.vcs_order,
                                     obj.subfile_order,
                                     obj.flagged,
                                     obj.rec_number,
                                     obj.rec_slot_number))

        return rf_inputs


#!/usr/bin/env python
#
# timestep: class representing a single timestep in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from .mwalib import CTimeStepS, CMetafitsMetadataS


class TimeStep:
    """
    A class representing a single TimeStep

    Attributes
    ----------
    index : int
        Ordinal index of this time step.

    unix_time_ms : int
        The UNIX time (in milliseconds) of the start of this time step

    gps_time_ms : int
        The GPS time (in milliseconds) of the start of this time step

    """

    def __init__(self,
                 index: int,
                 unix_time_ms: int,
                 gps_time_ms: int):
        """Initialise the class"""
        self.index: int = index
        self.unix_time_ms: int = unix_time_ms
        self.gps_time_ms: int = gps_time_ms

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Order: {self.index}, " \
               f"UNIX time: {float(self.unix_time_ms) / 1000.}, " \
               f"GPS time: {float(self.gps_time_ms) / 1000.})"

    @staticmethod
    def get_metafits_timesteps(metafits_metadata: CMetafitsMetadataS) -> []:
        """Retrieve all of the metafits timestep metadata and populate a list of metafits timesteps."""
        timesteps = []

        for i in range(0, metafits_metadata.num_metafits_timesteps):
            obj: CTimeStepS = metafits_metadata.metafits_timesteps[i]

            # Populate all the fields
            timesteps.append(TimeStep(i,
                                      obj.unix_time_ms,
                                      obj.gps_time_ms))

        return timesteps

    @staticmethod
    def get_correlator_or_voltage_timesteps(correlator_or_voltage_metadata) -> []:
        """Retrieve all of the timestep metadata and populate a list of timesteps."""
        timesteps = []

        for i in range(0, correlator_or_voltage_metadata.num_timesteps):
            obj: CTimeStepS = correlator_or_voltage_metadata.timesteps[i]

            # Populate all the fields
            timesteps.append(TimeStep(i,
                                      obj.unix_time_ms,
                                      obj.gps_time_ms))

        return timesteps


#!/usr/bin/env python
#
# Main class for access metafits metadata
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import ctypes as ct
from datetime import datetime
from .mwalib import mwalib_library,CMetafitsMetadataS, CMetafitsContextS,CCorrelatorContextS,CVoltageContextS,create_string_buffer
from .common import ERROR_MESSAGE_LEN, MWAMode, GeometricDelaysApplied
from .errors import PymwalibMetafitsMetadataGetError
from .antenna import Antenna
from .baseline import Baseline
from .rfinput import RFInput
from .coarse_channel import CoarseChannel
from .timestep import TimeStep

class MetafitsMetadata:
    def __init__(self,
                 metafits_context: ct.POINTER(CMetafitsContextS),
                 correlator_context: ct.POINTER(CCorrelatorContextS),
                 voltage_context: ct.POINTER(CVoltageContextS)):
        """Retrieve all of the metafits metadata and populate this class."""
        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

        c_object_ptr = ct.POINTER(CMetafitsMetadataS)()

        if mwalib_library.mwalib_metafits_metadata_get(metafits_context,
                                               correlator_context,
                                               voltage_context,
                                               ct.byref(c_object_ptr),
                                               error_message,
                                               ERROR_MESSAGE_LEN) != 0:
            raise PymwalibMetafitsMetadataGetError(
                f"Error creating metafits metadata object: {error_message.decode('utf-8').rstrip()}")
        else:
            # Populate all the fields
            c_object = c_object_ptr.contents

            self.obs_id: int = c_object.obs_id
            self.global_analogue_attenuation_db: float = c_object.global_analogue_attenuation_db
            self.ra_tile_pointing_deg: float = c_object.ra_tile_pointing_deg
            self.dec_tile_pointing_deg: float = c_object.dec_tile_pointing_deg
            self.ra_phase_center_deg: float = c_object.ra_phase_center_deg
            self.dec_phase_center_deg: float = c_object.dec_phase_center_deg
            self.az_deg: float = c_object.az_deg
            self.alt_deg: float = c_object.alt_deg
            self.za_deg: float = c_object.za_deg
            self.az_rad: float = c_object.az_rad
            self.alt_rad: float = c_object.alt_rad
            self.za_rad: float = c_object.za_rad
            self.sun_alt_deg: float = c_object.sun_alt_deg
            self.sun_distance_deg: float = c_object.sun_distance_deg
            self.moon_distance_deg: float = c_object.moon_distance_deg
            self.jupiter_distance_deg: float = c_object.jupiter_distance_deg
            self.lst_deg: float = c_object.lst_deg
            self.lst_rad: float = c_object.lst_rad
            self.hour_angle_string: str = c_object.hour_angle_string.decode("utf-8")
            self.grid_name: str = c_object.grid_name.decode("utf-8")
            self.grid_number: int = c_object.grid_number
            self.creator: str = c_object.creator.decode("utf-8")
            self.project_id: str = c_object.project_id.decode("utf-8")
            self.obs_name: str = c_object.obs_name.decode("utf-8")
            self.mode: MWAMode = c_object.mode
            self.geometric_delays_applied: GeometricDelaysApplied = c_object.geometric_delays_applied
            self.cable_delays_applied: bool = c_object.cable_delays_applied
            self.calibration_delays_and_gains_applied: bool = c_object.calibration_delays_and_gains_applied
            self.corr_fine_chan_width_hz: int = c_object.corr_fine_chan_width_hz
            self.corr_int_time_ms: int = c_object.corr_int_time_ms
            self.num_corr_fine_chans_per_coarse: int = c_object.num_corr_fine_chans_per_coarse
            self.sched_start_utc: datetime = datetime.utcfromtimestamp(c_object.sched_start_utc)
            self.sched_end_utc: datetime = datetime.utcfromtimestamp(c_object.sched_end_utc)
            self.sched_start_mjd: float = c_object.sched_start_mjd
            self.sched_end_mjd: float = c_object.sched_end_mjd
            self.sched_start_unix_time_ms: int = c_object.sched_start_unix_time_ms
            self.sched_end_unix_time_ms: int = c_object.sched_end_unix_time_ms
            self.sched_start_gps_time_ms: int = c_object.sched_start_gps_time_ms
            self.sched_end_gps_time_ms: int = c_object.sched_end_gps_time_ms
            self.sched_duration_ms: int = c_object.sched_duration_ms
            self.quack_time_duration_ms: int = c_object.quack_time_duration_ms
            self.good_time_unix_ms: int = c_object.good_time_unix_ms
            self.good_time_gps_ms: int = c_object.good_time_gps_ms
            self.num_ants: int = c_object.num_ants
            self.num_rf_inputs: int = c_object.num_rf_inputs
            self.num_ant_pols: int = c_object.num_ant_pols
            self.num_baselines: int = c_object.num_baselines
            self.num_visibility_pols: int = c_object.num_visibility_pols
            self.num_metafits_timesteps: int = c_object.num_metafits_timesteps
            self.num_metafits_coarse_chans: int = c_object.num_metafits_coarse_chans
            self.obs_bandwidth_hz: int = c_object.obs_bandwidth_hz
            self.coarse_chan_width_hz: int = c_object.coarse_chan_width_hz
            self.centre_freq_hz: int = c_object.centre_freq_hz
            self.metafits_filename: str = c_object.metafits_filename.decode("utf-8")

            # Populate rf_inputs
            self.rf_inputs = RFInput.get_rf_inputs(c_object)

            # Populate antennas
            self.antennas = Antenna.get_antennas(c_object, self.rf_inputs)

            # Populate baselines
            self.baselines = Baseline.get_baselines(c_object)

            # Populate timesteps
            self.metafits_timesteps = TimeStep.get_metafits_timesteps(c_object)

            # Populate coarse channels
            self.metafits_coarse_chans = CoarseChannel.get_metafits_coarse_channels(c_object)

            # We're now finished with the C memory, so free it
            mwalib_library.mwalib_metafits_metadata_free(c_object)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)

    def __str__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(\n" \
               f"Obs ID                                : {self.obs_id}\n" \
               f"Global attenuation                    : {self.global_analogue_attenuation_db} dB\n" \
               f"RA  (tile pointing)                   : {self.ra_tile_pointing_deg} deg\n" \
               f"Dec (tile pointing)                   : {self.dec_tile_pointing_deg} deg\n" \
               f"RA  (phase centre)                    : {self.ra_phase_center_deg} deg\n" \
               f"Dec (phase centre)                    : {self.dec_phase_center_deg} deg\n" \
               f"Az                                    : {self.az_deg} deg\n" \
               f"Alt                                   : {self.alt_deg} deg\n" \
               f"Zenith Angle                          : {self.za_deg} deg\n" \
               f"Az                                    : {self.az_rad} rad\n" \
               f"Alt                                   : {self.alt_rad} rad\n" \
               f"Zenith Angle                          : {self.za_rad} rad\n" \
               f"Sun Alt                               : {self.sun_alt_deg} deg\n" \
               f"Sun distance                          : {self.sun_distance_deg} deg\n" \
               f"Moon distance                         : {self.moon_distance_deg} deg\n" \
               f"Jupiter distance                      : {self.jupiter_distance_deg} deg\n" \
               f"LST                                   : {self.lst_deg} deg\n" \
               f"LST                                   : {self.lst_rad} rad\n" \
               f"HA                                    : {self.hour_angle_string} dms\n" \
               f"Grid name                             : {self.grid_name}\n" \
               f"Grid number                           : {self.grid_number}\n" \
               f"Creator                               : {self.creator}\n" \
               f"Project ID                            : {self.project_id}\n" \
               f"Observation Name                      : {self.obs_name}\n" \
               f"Mode                                  : {self.mode}\n" \
               f"geometric_delays_applied              : {self.geometric_delays_applied}\n" \
               f"cable_delays_applied                  : {self.cable_delays_applied}\n"\
               f"calibration_delays_and_gains_applied  : {self.calibration_delays_and_gains_applied}\n" \
               f"Correlator fine channel width (Hz)    : {self.corr_fine_chan_width_hz}\n" \
               f"Correlator int. time (ms)             : {self.corr_int_time_ms}\n" \
               f"Correlator fine channels per coarse   : {self.num_corr_fine_chans_per_coarse}\n" \
               f"Scheduled Start (UTC)                 : {self.sched_start_utc}\n" \
               f"Scheduled End (UTC)                   : {self.sched_end_utc}\n" \
               f"Scheduled Start (UNIX)                : {float(self.sched_start_unix_time_ms) / 1000.} UNIX\n" \
               f"Scheduled End (UNIX)                  : {float(self.sched_end_unix_time_ms) / 1000.} UNIX\n" \
               f"Scheduled Start (MJD)                 : {self.sched_start_mjd} MJD\n" \
               f"Scheduled End (MJD)                   : {self.sched_end_mjd} MJD\n" \
               f"Scheduled Duration                    : {float(self.sched_duration_ms) / 1000.} s\n" \
               f"Quack time (ms)                       : {float(self.quack_time_duration_ms) / 1000.} s\n" \
               f"Good start time (UNIX)                : {float(self.good_time_unix_ms) / 1000.} UNIX\n" \
               f"Good start time (GPS)                 : {float(self.good_time_gps_ms) / 1000.} GPS\n" \
               f"num antennas                          : {self.num_ants}\n" \
               f"num rf_inputs                         : {self.num_rf_inputs}\n" \
               f"Baselines                             : {self.num_baselines}\n" \
               f"Visibility pols                       : {self.num_visibility_pols}\n" \
               f"num antenna pols                      : {self.num_ant_pols}\n" \
               f"num metafits timesteps                : {self.num_metafits_timesteps}\n" \
               f"num metafits coarse channe            : {self.num_metafits_coarse_chans}\n" \
               f"Observation bandwidth                 : {float(self.obs_bandwidth_hz) / 1000000.} MHz\n" \
               f"Coarse channel width                  : {float(self.coarse_chan_width_hz) / 1000000.} MHz\n)\n"



#!/usr/bin/env python
#
# Version handling for pymwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from .mwalib import mwalib_library
from .errors import PymwalibMwalibVersionNotCompatibleError

"""Returns the major, minor and patch version of pymwalib as a string"""
def get_pymwalib_version_string() -> str:
    import pkg_resources  # part of setuptools
    package_version = pkg_resources.require("pymwalib")[0].version
    return package_version

"""Returns the major, minor and patch version of mwalib as a string"""
def get_mwalib_version_string() -> str:
    mwalib_major = mwalib_library.mwalib_get_version_major()
    mwalib_minor = mwalib_library.mwalib_get_version_minor()
    mwalib_patch = mwalib_library.mwalib_get_version_patch()

    return f"{mwalib_major}.{mwalib_minor}.{mwalib_patch}"

"""Returns a the major, minor and patch version of pymwalib"""
def get_pymwalib_version_number() -> (int, int, int):
    version = get_pymwalib_version_string()
    try:
        return int(version.split(".")[0]),int(version.split(".")[1]), int(version.split(".")[2])
    except Exception as e:
        raise Exception(f"Unabled to determine pymwalib version: Got {version} which could not be parsed. Error: {e}")

"""Checks, using semantic versioning, that mwalib is compatible with pymwalib. If so, returns, otherwise raises an execption"""
def check_mwalib_version():
    # Perform a version check before going too far
    mwalib_major = mwalib_library.mwalib_get_version_major()
    mwalib_minor = mwalib_library.mwalib_get_version_minor()
    mwalib_patch = mwalib_library.mwalib_get_version_patch()

    pymwalib_version = get_pymwalib_version_number()
    pymwalib_major = pymwalib_version[0]
    pymwalib_minor = pymwalib_version[1]
    pymwalib_patch = pymwalib_version[2]

    # Use semantic rules to determine compatibility
    # if major versions don't match then we are not compatible
    if mwalib_major != pymwalib_major:
        raise PymwalibMwalibVersionNotCompatibleError(f"pymwalib version {pymwalib_major}.{pymwalib_minor}.* is not "
                                                 f"compatible with mwalib {mwalib_major}.{mwalib_minor}.*")

    # pre release rules also apply to major version 0
    # if ANY part of versions don't match then we are not compatible
    if mwalib_major == 0:
        if mwalib_minor != pymwalib_minor or mwalib_patch != pymwalib_patch:
            raise PymwalibMwalibVersionNotCompatibleError(f"pymwalib version "
                                                     f"{pymwalib_major}.{pymwalib_minor}.{pymwalib_patch} is not "
                                                     f"compatible with mwalib "
                                                     f"{mwalib_major}.{mwalib_minor}.{mwalib_patch}")



#!/usr/bin/env python
#
# CorrelatorContext: main interface for pymwalib for correlator data
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import numpy.ctypeslib as npct
import ctypes as ct
from .mwalib import CCorrelatorContextS, CCorrelatorMetadataS, mwalib_library,create_string_buffer,MWALIB_SUCCESS,MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN
from .coarse_channel import CoarseChannel
from .common import ERROR_MESSAGE_LEN, MWAVersion
from .errors import PymwalibCorrelatorContextNewError,PymwalibCorrelatorContextDisplayError,PymwalibNoDataForTimestepAndCoarseChannelError,PymwalibCorrelatorContextReadByBaselineError,PymwalibCorrelatorContextReadByFrequencyError, PymwalibCorrelatorMetadataGetError
from .metafits_metadata import MetafitsMetadata
from .timestep import TimeStep
from .version import check_mwalib_version


class CorrelatorContext:
    """Main class to interface with mwalib correlator observations"""
    def __init__(self, metafits_filename: str, gpubox_filenames: list):
        """Take metafits and gpubox files, and populate this class via mwalib"""
        #
        # Ensure we have a compatible version of mwalib
        #
        check_mwalib_version()

        error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)
        self._correlator_context_object = ct.POINTER(CCorrelatorContextS)()

        # First populate the context object
        self._get_correlator_context(metafits_filename, gpubox_filenames)

        # Get correlator metadata
        c_object_ptr = ct.POINTER(CCorrelatorMetadataS)()
        if mwalib_library.mwalib_correlator_metadata_get(self._correlator_context_object,
                                                         ct.byref(c_object_ptr),
                                                         error_message,
                                                         ERROR_MESSAGE_LEN) != 0:
            raise PymwalibCorrelatorMetadataGetError(
                f"Error creating correlator metadata object: {error_message.decode('utf-8').rstrip()}")

        c_object = c_object_ptr.contents

        # Populate all the fields
        self.mwa_version: MWAVersion = MWAVersion(c_object.mwa_version)

        self.num_timesteps: int = c_object.num_timesteps
        self.num_coarse_chans: int = c_object.num_coarse_chans

        self.num_common_timesteps: int = c_object.num_common_timesteps
        self.num_common_coarse_chans: int = c_object.num_common_coarse_chans
        self.common_start_unix_time_ms = c_object.common_start_unix_time_ms
        self.common_end_unix_time_ms = c_object.common_end_unix_time_ms
        self.common_start_gps_time_ms = c_object.common_start_gps_time_ms
        self.common_end_gps_time_ms = c_object.common_end_gps_time_ms
        self.common_duration_ms = c_object.common_duration_ms
        self.common_bandwidth_hz = c_object.common_bandwidth_hz

        self.num_common_good_timesteps: int = c_object.num_common_good_timesteps
        self.num_common_good_coarse_chans: int = c_object.num_common_good_coarse_chans
        self.common_good_start_unix_time_ms = c_object.common_good_start_unix_time_ms
        self.common_good_end_unix_time_ms = c_object.common_good_end_unix_time_ms
        self.common_good_start_gps_time_ms = c_object.common_good_start_gps_time_ms
        self.common_good_end_gps_time_ms = c_object.common_good_end_gps_time_ms
        self.common_good_duration_ms = c_object.common_good_duration_ms
        self.common_good_bandwidth_hz = c_object.common_good_bandwidth_hz

        self.num_provided_timesteps: int = c_object.num_provided_timesteps
        self.num_provided_coarse_chans: int = c_object.num_provided_coarse_chans

        self.num_timestep_coarse_chan_bytes: int = c_object.num_timestep_coarse_chan_bytes
        self.num_timestep_coarse_chan_floats: int = c_object.num_timestep_coarse_chan_floats
        self.num_gpubox_files: int = c_object.num_gpubox_files

        # Common timesteps
        self.common_timestep_indices = []
        for i in range(self.num_common_timesteps):
            self.common_timestep_indices.append(c_object.common_timestep_indices[i])

        # Common coarse chans
        self.common_coarse_chan_indices = []
        for i in range(self.num_common_coarse_chans):
            self.common_coarse_chan_indices.append(c_object.common_coarse_chan_indices[i])

        # Common Good timesteps
        self.common_good_timestep_indices = []
        for i in range(self.num_common_good_timesteps):
            self.common_good_timestep_indices.append(c_object.common_good_timestep_indices[i])

        # Common Good coarse chans
        self.common_good_coarse_chan_indices = []
        for i in range(self.num_common_good_coarse_chans):
            self.common_good_coarse_chan_indices.append(c_object.common_good_coarse_chan_indices[i])

        # Provided timesteps
        self.provided_timestep_indices = []
        for i in range(self.num_provided_timesteps):
            self.provided_timestep_indices.append(c_object.provided_timestep_indices[i])

        # Provided coarse chans
        self.provided_coarse_chan_indices = []
        for i in range(self.num_provided_coarse_chans):
            self.provided_coarse_chan_indices.append(c_object.provided_coarse_chan_indices[i])

        # Now Get metafits context
        self.metafits_context: MetafitsMetadata = MetafitsMetadata(None, self._correlator_context_object, None)

        # Populate coarse channels
        self.coarse_channels = CoarseChannel.get_correlator_or_voltage_coarse_channels(c_object_ptr.contents)

        # Populate timesteps
        self.timesteps = TimeStep.get_correlator_or_voltage_timesteps(c_object_ptr.contents)

        # We're now finished with the C memory, so free it
        mwalib_library.mwalib_correlator_metadata_free(c_object)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Ensure we free the rust allocated memory in mwalib"""
        if self._correlator_context_object:
            mwalib_library.mwalib_correlator_context_free(self._correlator_context_object)

    def _get_correlator_context(self, metafits_filename: str, gpubox_filenames: list):
        """This method will read and validate the metafits and gpubox files. If all has worked, then
        the context object can be used in subsequent calls to populate aspects of this class."""
        if mwalib_library:
            # Encode all inputs as UTF-8.
            m = ct.c_char_p(metafits_filename.encode("utf-8"))

            # https://stackoverflow.com/questions/4145775/how-do-i-convert-a-python-list-into-a-c-array-by-using-ctypes
            encoded = []
            for g in gpubox_filenames:
                encoded.append(ct.c_char_p(g.encode("utf-8")))
            seq = ct.c_char_p * len(encoded)
            g = seq(*encoded)
            error_message: bytes = create_string_buffer(ERROR_MESSAGE_LEN)

            if mwalib_library.mwalib_correlator_context_new(
                m, g, len(encoded), ct.byref(self._correlator_context_object), error_message, ERROR_MESSAGE_LEN) != 0:
                raise PymwalibCorrelatorContextNewError(f"Error creating correlator context object: "
                                             f"{error_message.decode('utf-8').rstrip()}")
        else:
            raise PymwalibCorrelatorContextNewError(f"Error creating correlator context object: mwalib is not loaded.")

    def display(self):
         """Displays a human readable summary of the correlator context"""
         error_message = create_string_buffer(ERROR_MESSAGE_LEN)

         if mwalib_library.mwalib_correlator_context_display(self._correlator_context_object,
                                                     error_message,
                                                     ERROR_MESSAGE_LEN) != 0:
             raise PymwalibCorrelatorContextDisplayError(f"Error calling mwalib_correlator_context_display(): "
                                                        f"{error_message.decode('utf-8').rstrip()}")

    def read_by_baseline(self, timestep_index: int, coarse_chan_index: int):
         """Retrieve one HDU (ordered baseline,freq,pol,r,i) as a numpy array."""
         error_message = " ".encode("utf-8") * ERROR_MESSAGE_LEN

         float_buffer_type = ct.c_float * self.num_timestep_coarse_chan_floats
         buffer = float_buffer_type()

         ret_val = mwalib_library.mwalib_correlator_context_read_by_baseline(self._correlator_context_object,
                                                              ct.c_size_t(timestep_index),
                                                              ct.c_size_t(coarse_chan_index),
                                                              buffer,
                                                              self.num_timestep_coarse_chan_floats,
                                                              error_message, ERROR_MESSAGE_LEN)

         if ret_val == MWALIB_SUCCESS:
             return npct.as_array(buffer, shape=(self.num_timestep_coarse_chan_floats,))
         elif ret_val == MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN:
             raise PymwalibNoDataForTimestepAndCoarseChannelError(f"No data exists for this timestep {timestep_index} and coarse channel {coarse_chan_index}")
         else:
             raise PymwalibCorrelatorContextReadByBaselineError(f"Error reading data: "
                                                  f"{error_message.decode('utf-8').rstrip()}")


    def read_by_frequency(self, timestep_index: int, coarse_chan_index: int):
        """Retrieve one HDU (ordered freq,baseline,pol,r,i) as a numpy array."""
        error_message = " ".encode("utf-8") * ERROR_MESSAGE_LEN

        float_buffer_type = ct.c_float * self.num_timestep_coarse_chan_floats
        buffer = float_buffer_type()

        ret_val = mwalib_library.mwalib_correlator_context_read_by_frequency(self._correlator_context_object,
                                                              ct.c_size_t(timestep_index),
                                                              ct.c_size_t(coarse_chan_index),
                                                              buffer,
                                                              self.num_timestep_coarse_chan_floats,
                                                              error_message, ERROR_MESSAGE_LEN)
        if ret_val == MWALIB_SUCCESS:
            return npct.as_array(buffer, shape=(self.num_timestep_coarse_chan_floats,))
        elif ret_val == MWALIB_NO_DATA_FOR_TIMESTEP_COARSECHAN:
            raise PymwalibNoDataForTimestepAndCoarseChannelError(f"No data exists for this timestep {timestep_index} and coarse channel {coarse_chan_index}")
        else:
            raise PymwalibCorrelatorContextReadByFrequencyError(f"Error reading data: "
                                                                   f"{error_message.decode('utf-8').rstrip()}")
    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(\n" \
               f"MWA Version                           : {MWAVersion(self.mwa_version).name}\n" \
               f"Num timesteps                         : {self.num_timesteps}\n" \
               f"Num coarse channels                   : {self.num_coarse_chans}\n" \
               f"num_provided_timestep_indices         : {self.num_provided_timesteps}\n" \
               f"num_provided_coarse_chan_indices      : {self.num_provided_coarse_chans}\n" \
               f"(common) Start time (UNIX)            : {float(self.common_start_unix_time_ms) / 1000.}\n" \
               f"(common) End time (UNIX)              : {float(self.common_end_unix_time_ms) / 1000.}\n" \
               f"(common) Start time (GPS)             : {float(self.common_start_gps_time_ms) / 1000.}\n" \
               f"(common) End time (GPS)               : {float(self.common_end_gps_time_ms) / 1000.}\n" \
               f"(common) Duration                     : {float(self.common_duration_ms) / 1000.} s\n" \
               f"(common) num timesteps                : {self.num_common_timesteps}\n" \
               f"(common) num coarse channels          : {self.num_common_coarse_chans}\n" \
               f"(common) Bandwidth                    : {float(self.common_bandwidth_hz) / 1000000.} MHz\n" \
               f"(common good) Start time (UNIX)       : {float(self.common_good_start_unix_time_ms) / 1000.}\n" \
               f"(common good) End time (UNIX)         : {float(self.common_good_end_unix_time_ms) / 1000.}\n" \
               f"(common good) Start time (GPS)        : {float(self.common_good_start_gps_time_ms) / 1000.}\n" \
               f"(common good) End time (GPS)          : {float(self.common_good_end_gps_time_ms) / 1000.}\n" \
               f"(common good) Duration                : {float(self.common_good_duration_ms) / 1000.} s\n" \
               f"(common good) num timesteps           : {self.num_common_good_timesteps}\n" \
               f"(common good) num coarse channels     : {self.num_common_good_coarse_chans}\n" \
               f"(common good) Bandwidth               : {float(self.common_good_bandwidth_hz) / 1000000.} MHz\n" \
               f"Num bytes per timestep coarse channel : " \
               f"{float(self.num_timestep_coarse_chan_bytes) / (1024. * 1024.)} MB\n" \
               f"Num floats per timestep coarse channel: " \
               f"{self.num_timestep_coarse_chan_floats}\n" \
               f"(actual) num GPUBox files             : {self.num_gpubox_files}\n)\n"


#!/usr/bin/env python
#
# pymwalib examples/print-context - run through all of pymwalib's objects
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import argparse
import numpy as np
from pymwalib.common import MWAVersion
from pymwalib.metafits_context import MetafitsContext
from pymwalib.correlator_context import CorrelatorContext
from pymwalib.voltage_context import VoltageContext
from pymwalib.version import check_mwalib_version

if __name__ == "__main__":
    # ensure we have a compatible mwalib first
    # You can skip this if you want, but your first pymwalib call will raise an error. Best trap it here
    # and provide a nice user message
    try:
        check_mwalib_version()
    except Exception as e:
        print(e)
        exit(1)

    parser = argparse.ArgumentParser()

    parser.add_argument("-m", "--metafits", required=True,
                        help="Path to the metafits file.")
    parser.add_argument("datafiles", nargs='*',
                        help="Paths to the gpubox or voltage data files (or neither).")
    args = parser.parse_args()

    context = None

    # Check what we have for the data files
    if len(args.datafiles) == 0:
        # We invoke a metafits context
        print("Only metafits file provided, assuming Legacy Correlator interpretation of metafits.")
        context = MetafitsContext(args.metafits, MWAVersion.CorrLegacy)
    else:
        corr_suffixes = [x for x in args.datafiles if x[-5:] == ".fits"]
        dat_suffixes = [x for x in args.datafiles if x[-4:] == ".dat"]
        sub_suffixes = [x for x in args.datafiles if x[-4:] == ".sub"]

        if len(corr_suffixes) + len(dat_suffixes) + len(sub_suffixes) == 0:
            print("Error- no .fits, .dat or .sub files provided")
            exit(-2)
        elif len(corr_suffixes) > 0 and len(dat_suffixes) + len(sub_suffixes) == 0:
            print(f"{len(corr_suffixes)} correlator/gpubox files detected")
            context = CorrelatorContext(args.metafits, args.datafiles)
        elif (len(dat_suffixes) > 0 and len(corr_suffixes) + len(sub_suffixes) == 0) or \
             (len(sub_suffixes) > 0 and len(corr_suffixes) + len(dat_suffixes) == 0):
            print(f"{len(dat_suffixes)} voltage data files detected")
            context = VoltageContext(args.metafits, args.datafiles)
        else:
            print("Error: Combination of different data files supplied.")
            exit(-3)

    # Test the debug "display" method
    print("\nTesting Display method:")
    context.display()



from distutils.core import setup
import setuptools

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'pymwalib',
  packages = ['pymwalib'],
  version = '0.8.4',
  license = 'mpl-2.0',
  description = 'A Python interface for mwalib, a library to read Murchison Widefield Array (MWA) raw visibilities, voltages and metadata into a common structure',
  long_description = long_description,
  long_description_content_type='text/markdown',
  author = 'Greg Sleap',
  author_email = 'greg.sleap@curtin.edu.au',
  url = 'https://github.com/MWATelescope/pymwalib',
  download_url = 'https://github.com/MWATelescope/pymwalib/archive/v0.6.2.tar.gz',
  keywords = ['MWA', 'radioastronomy'],
  install_requires = [
          'numpy',
      ],
  classifiers = [
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Astronomy',
    'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
  ],
)




#!/usr/bin/env python
#
# coarse_channel: class representing a single coarse_channel in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from .mwalib import CCoarseChannelS, CMetafitsMetadataS


class CoarseChannel:
    """
    A class representing a single CoarseChannel

    Attributes
    ----------
    index : int
        Ordinal index of this coarse channel.

    Please see https://github.com/MWATelescope/mwalib/blob/master/src/coarse_channel.rs for remaining attributes

    """

    def __init__(self,
                 index: int,
                 corr_chan_number: int,
                 rec_chan_number: int,
                 gpubox_number: int,
                 chan_width_hz: int,
                 chan_start_hz: int,
                 chan_centre_hz: int,
                 chan_end_hz: int):
        """Initialise the class"""
        self.index: int = index
        self.corr_chan_number: int = corr_chan_number
        self.rec_chan_number: int = rec_chan_number
        self.gpubox_number: int = gpubox_number
        self.chan_width_hz: int = chan_width_hz
        self.chan_start_hz: int = chan_start_hz
        self.chan_centre_hz: int = chan_centre_hz
        self.chan_end_hz: int = chan_end_hz

    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Order: {self.index}, " \
               f"Correlator channel number: {self.corr_chan_number}, " \
               f"Receiver channel number: {self.rec_chan_number}, " \
               f"GPUBox Number: {self.gpubox_number}, " \
               f"Channel width MHz: {float(self.chan_width_hz) / 1000000.}, " \
               f"Channel start MHz: {float(self.chan_start_hz) / 1000000.}, " \
               f"Channel centre MHz: {float(self.chan_centre_hz) / 1000000.}, " \
               f"Channel end MHz: {float(self.chan_end_hz) / 1000000.})"

    @staticmethod
    def get_metafits_coarse_channels(metafits_metadata: CMetafitsMetadataS) -> []:
        """Retrieve all of the coarse_channel metadata and populate a list of coarse_channels."""
        coarse_channels = []

        for i in range(0, metafits_metadata.num_metafits_coarse_chans):
            obj: CCoarseChannelS = metafits_metadata.metafits_coarse_chans[i]

            # Populate all the fields
            coarse_channels.append(CoarseChannel(i,
                                                 obj.corr_chan_number,
                                                 obj.rec_chan_number,
                                                 obj.gpubox_number,
                                                 obj.chan_width_hz,
                                                 obj.chan_start_hz,
                                                 obj.chan_centre_hz,
                                                 obj.chan_end_hz, ))

        return coarse_channels

    @staticmethod
    def get_correlator_or_voltage_coarse_channels(correlator_or_voltage_metadata) -> []:
        """Retrieve all of the coarse_channel metadata and populate a list of coarse_channels."""
        coarse_channels = []

        for i in range(0, correlator_or_voltage_metadata.num_coarse_chans):
            obj: CCoarseChannelS = correlator_or_voltage_metadata.coarse_chans[i]

            # Populate all the fields
            coarse_channels.append(CoarseChannel(i,
                                                 obj.corr_chan_number,
                                                 obj.rec_chan_number,
                                                 obj.gpubox_number,
                                                 obj.chan_width_hz,
                                                 obj.chan_start_hz,
                                                 obj.chan_centre_hz,
                                                 obj.chan_end_hz, ))

        return coarse_channels


#!/usr/bin/env python
#
# antenna: class representing a single antenna in mwalib
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from .mwalib import  CAntennaS, CMetafitsMetadataS
from .rfinput import RFInput

class Antenna:
    """
    A class representing a single mwalibAntenna

    Attributes
    ----------
    index : int
        Ordinal index of this RF Input.

    Please see https://github.com/MWATelescope/mwalib/blob/master/src/antenna.rs for remaining attributes

    """

    def __init__(self,
                 index: int,
                 ant: int,
                 tile_id: int,
                 tile_name: str,
                 rf_input_x: RFInput,
                 rf_input_y: RFInput,
                 electrical_length_m: float,
                 north_m: float,
                 east_m: float,
                 height_m: float):
        """Initialise the class"""
        self.index: int = index
        self.ant: int = ant
        self.tile_id: int = tile_id
        self.tile_name: str = tile_name
        self.rf_input_x: RFInput = rf_input_x
        self.rf_input_y: RFInput = rf_input_y
        self.electrical_length_m = electrical_length_m
        self.north_m = north_m
        self.east_m = east_m
        self.height_m = height_m


    def __repr__(self):
        """Returns a representation of the class"""
        return f"{self.__class__.__name__}(" \
               f"Index: {self.index}, " \
               f"ant: {self.ant}, " \
               f"Tile Id: {self.tile_id}, " \
               f"Tile Name: {self.tile_name}, " \
               f"RF Input X: {self.rf_input_x!r}, " \
               f"RF Input Y: {self.rf_input_y!r}, " \
               f"Electrical length: {self.electrical_length_m!r} m, " \
               f"North: {self.north_m} m, " \
               f"East: {self.east_m} m, " \
               f"Height: {self.height_m} m)"

    @staticmethod
    def get_antennas(metafits_metadata: CMetafitsMetadataS,
                     rf_inputs: []) -> []:
        """Retrieve all of the antenna metadata and populate a list of antennas."""
        antennas = []

        for i in range(0, metafits_metadata.num_ants):
            obj: CAntennaS = metafits_metadata.antennas[i]

            # Populate all the fields
            antennas.append(Antenna(i,
                             obj.ant,
                             obj.tile_id,
                             obj.tile_name.decode("utf-8"),
                             rf_inputs[obj.rfinput_x],
                             rf_inputs[obj.rfinput_y],
                             obj.electrical_length_m,
                             obj.north_m,
                             obj.east_m,
                             obj.height_m))

        return antennas